/*-
 * Copyright (c) 2004 - 2009 CTPP Team
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the CTPP Team nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      VMSTDLibFunctions.cpp
 *
 * $CTPP$
 */

#include <VMSTDLibFunctions.hpp>

#include <CDT.hpp>
#include <CTPP2Util.hpp>
#include <OutputCollector.hpp>
#include <StaticData.hpp>
#include <StaticText.hpp>
#include <SyscallFactory.hpp>

#include <errno.h>
#include <strings.h>
#include <sys/time.h>
#include <time.h>

#ifdef GETTEXT_SUPPORT
#include <libintl.h>
#endif // GETTEXT_SUPPORT

#ifdef MD5_SUPPORT
    #ifdef MD5_WITHOUT_OPENSSL
        #include <md5.h>
        #define MD5_Init    MD5Init
        #define MD5_Update  MD5Update
        #define MD5_Final   MD5Final
    #else
        #include <openssl/md5.h>
    #endif
#endif // MD5_SUPPORT

namespace CTPP // C++ Template Engine
{

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnEmitter
//

//
// Constructor
//
FnEmitter::FnEmitter(): pCDT(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnEmitter::PreExecuteSetup(OutputCollector          & oCollector,
                                  CDT                      & oCDT,
                                  const ReducedStaticText  & oSyscalls,
                                  const ReducedStaticData  & oStaticData,
                                  const ReducedStaticText  & oStaticText)
{
	pCDT = &oCDT;
return 0;
}

//
// Handler
//
INT_32 FnEmitter::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	oCDTRetVal = *pCDT;
return 0;
}

//
// Get function name
//
CCHAR_P FnEmitter::GetName() { return CTPP2_INT_HANDLER_PREFIX "_emitter"; }

//
// A destructor
//
FnEmitter::~FnEmitter() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutput
//

//
// Constructor
//
FnOutput::FnOutput(): pCollector(NULL) { ;; }

//
// Pre-execution handler setup
//
INT_32 FnOutput::PreExecuteSetup(OutputCollector          & oCollector,
                                 CDT                      & oCDT,
                                 const ReducedStaticText  & oSyscalls,
                                 const ReducedStaticData  & oStaticData,
                                 const ReducedStaticText  & oStaticText)
{
	pCollector = &oCollector;
return 0;
}

//
// Handler
//
INT_32 FnOutput::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	for (INT_32 iI = iArgNum - 1; iI >= 0; --iI)
	{
		std::string sTMP(aArguments[iI].GetString());
		if (pCollector -> Collect(sTMP.c_str(), sTMP.length()) == -1) { return -1; }
	}
return 0;
}

//
// Get function name
//
CCHAR_P FnOutput::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output"; }

//
// A destructor
//
FnOutput::~FnOutput() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutputText
//

//
// Constructor
//
FnOutputText::FnOutputText(): pCollector(NULL),
                              pStaticText(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnOutputText::PreExecuteSetup(OutputCollector          & oCollector,
                                     CDT                      & oCDT,
                                     const ReducedStaticText  & oSyscalls,
                                     const ReducedStaticData  & oStaticData,
                                     const ReducedStaticText  & oStaticText)
{
	pCollector  = &oCollector;
	pStaticText = &oStaticText;
return 0;
}

//
// Handler
//
INT_32 FnOutputText::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1)
	{
		UINT_32 iDatalength = 0;
		CCHAR_P sTMP = pStaticText -> GetData(aArguments[0].GetInt(), iDatalength);
		if (pCollector -> Collect(sTMP, iDatalength) == -1) { return -1; }

		return 0;
	}
return -1;
}

//
// Get function name
//
CCHAR_P FnOutputText::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output_text"; }

//
// A destructor
//
FnOutputText::~FnOutputText() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutputInt
//

//
// Constructor
//
FnOutputInt::FnOutputInt(): pCollector(NULL),
                            pStaticData(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnOutputInt::PreExecuteSetup(OutputCollector          & oCollector,
                                    CDT                      & oCDT,
                                    const ReducedStaticText  & oSyscalls,
                                    const ReducedStaticData  & oStaticData,
                                    const ReducedStaticText  & oStaticText)
{
	pCollector  = &oCollector;
	pStaticData = &oStaticData;
return 0;
}

//
// Handler
//
INT_32 FnOutputInt::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	char sBuffer[2048 + 1];
	if (iArgNum == 1)
	{
		long long iData = pStaticData -> GetInt(aArguments[0].GetInt());

		INT_32 iChars = snprintf(sBuffer, 2048, "%lli", iData);
		if (iChars < 0) { return -1; }
		if (pCollector -> Collect(sBuffer, iChars) == -1) { return -1; }

		return 0;
	}
return -1;
}

//
// Get function name
//
CCHAR_P FnOutputInt::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output_int"; }

//
// A destructor
//
FnOutputInt::~FnOutputInt() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnOutputFloat
//

//
// Constructor
//
FnOutputFloat::FnOutputFloat(): pCollector(NULL),
                                pStaticData(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnOutputFloat::PreExecuteSetup(OutputCollector          & oCollector,
                                      CDT                      & oCDT,
                                      const ReducedStaticText  & oSyscalls,
                                      const ReducedStaticData  & oStaticData,
                                      const ReducedStaticText  & oStaticText)
{
	pCollector  = &oCollector;
	pStaticData = &oStaticData;
return 0;
}

//
// Handler
//
INT_32 FnOutputFloat::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	char sBuffer[2048 + 1];
	if (iArgNum == 1)
	{
		double dData = pStaticData -> GetFloat(aArguments[0].GetInt());

		INT_32 iChars = snprintf(sBuffer, 2048, "%f", dData);
		if (iChars < 0) { return -1; }
		if (pCollector -> Collect(sBuffer, iChars) == -1) { return -1; }

		return 0;
	}
return -1;
}

//
// Get function name
//
CCHAR_P FnOutputFloat::GetName() { return CTPP2_INT_HANDLER_PREFIX "_output_float"; }

//
// A destructor
//
FnOutputFloat::~FnOutputFloat() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnSprintf
//

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnURLEscape
//

//
// Constructor
//
FnURLEscape::FnURLEscape()
{
	;;
}

//
// Handler
//
INT_32 FnURLEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1)
	{
		oCDTRetVal = URLEscape(aArguments[0].GetString());
		return 0;
	}

return -1;
}

//
// Get function name
//
CCHAR_P FnURLEscape::GetName() { return "urlescape"; }

//
// A destructor
//
FnURLEscape::~FnURLEscape() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnHTMLEscape
//

//
// Constructor
//
FnHTMLEscape::FnHTMLEscape()
{
	;;
}

//
// Handler
//
INT_32 FnHTMLEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = HTMLEscape(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnHTMLEscape::GetName() { return "htmlescape"; }

//
// A destructor
//
FnHTMLEscape::~FnHTMLEscape() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnXMLEscape
//

//
// Constructor
//
FnXMLEscape::FnXMLEscape()
{
	;;
}

//
// Handler
//
INT_32 FnXMLEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = XMLEscape(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnXMLEscape::GetName() { return "xmlescape"; }

//
// A destructor
//
FnXMLEscape::~FnXMLEscape() throw() { ;; }


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnJSEscape
//

//
// Constructor
//
FnJSEscape::FnJSEscape()
{
	;;
}

//
// Handler
//
INT_32 FnJSEscape::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	const CDT & oCDT = aArguments[0];
	switch (oCDT.GetType())
	{
		case CDT::UNDEF:
			oCDTRetVal = "null";
			break;

		case CDT::INT_VAL:
		case CDT::REAL_VAL:
		case CDT::POINTER_VAL:
			oCDTRetVal = oCDT.GetString();
			break;

		case CDT::STRING_VAL:
			oCDTRetVal = "\"";
			oCDTRetVal.Append(EscapeJSONString(oCDT.GetString()));
			oCDTRetVal.Append("\"");
			break;
		default:
			oCDTRetVal = "";
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnJSEscape::GetName() { return "jsescape"; }

//
// A destructor
//
FnJSEscape::~FnJSEscape() throw() { ;; }


// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnGetText
//
#ifdef GETTEXT_SUPPORT

//
// Constructor
//
FnGetText::FnGetText()
{
	;;
}

//
// Handler
//
INT_32 FnGetText::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1)
	{
		// Get string to escape
		oCDTRetVal = gettext(aArguments[0].GetString().c_str());
		return 0;
	}
return -1;
}

//
// Get function name
//
CCHAR_P FnGetText::GetName() { return "gettext"; }

//
// A destructor
//
FnGetText::~FnGetText() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnGetText_
//

//
// Constructor
//
FnGetText_::FnGetText_()
{
	;;
}

//
// Get function name
//
CCHAR_P FnGetText_::GetName() { return "_"; }

//
// A destructor
//
FnGetText_::~FnGetText_() throw() { ;; }

#endif // GETTEXT_SUPPORT

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnInSet
//

//
// Constructor
//
FnInSet::FnInSet()
{
	;;
}

//
// Handler
//
INT_32 FnInSet::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1) { return -1; }

	INT_32 iI = iArgNum - 1;

	CDT & oCDT1 = aArguments[iI];

	--iI;
	for (; iI >= 0; --iI)
	{
		if (oCDT1 == aArguments[iI]) { oCDTRetVal = 1; return 0; }
	}
	oCDTRetVal = 0;
return 0;
}

//
// Get function name
//
CCHAR_P FnInSet::GetName() { return "in_set"; }

//
// A destructor
//
FnInSet::~FnInSet() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnHrefParam
//

//
// Constructor
//
FnHrefParam::FnHrefParam()
{
	;;
}

//
// Handler
//
INT_32 FnHrefParam::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 2) { return -1; }

	oCDTRetVal = URLEscape(aArguments[1].GetString());
	oCDTRetVal.Append("=");
	oCDTRetVal += URLEscape(aArguments[0].GetString());
	oCDTRetVal.Append("&");

return 0;
}

//
// Get function name
//
CCHAR_P FnHrefParam::GetName() { return "href_param"; }

//
// A destructor
//
FnHrefParam::~FnHrefParam() throw() { ;; }

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnFormParam
//

//
// Constructor
//
FnFormParam::FnFormParam()
{
	;;
}

//
// Handler
//
INT_32 FnFormParam::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 2) { return -1; }

	oCDTRetVal = "<input type=\"hidden\" name=\"";
	oCDTRetVal += HTMLEscape(aArguments[1].GetString());
	oCDTRetVal.Append("\" value=\"");
	oCDTRetVal += HTMLEscape(aArguments[0].GetString());
	oCDTRetVal.Append("\" />");

return 0;
}

//
// Get function name
//
CCHAR_P FnFormParam::GetName() { return "form_param"; }

//
// A destructor
//
FnFormParam::~FnFormParam() throw() { ;; }


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnDateFormat
//

//
// Constructor
//
FnDateFormat::FnDateFormat()
{
	;;
}

//
// Handler
//
INT_32 FnDateFormat::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Only 2 arguments
	if (iArgNum != 2) { return -1; }
	// Temp
	CHAR_8 szBuffer[CTPP_ESCAPE_BUFFER_LEN + 1];

	CDT & oTMP = aArguments[1];
	time_t iTime = 0;
	if      (oTMP.GetType() <= CDT::REAL_VAL)   { iTime = oTMP.GetInt(); }
	else if (oTMP.GetType() == CDT::STRING_VAL)
	{
		long long iLL = 0;
		sscanf(oTMP.GetString().c_str(), "%lld", &iLL);
		iTime = INT_64(iLL);
	}

	const struct tm * pTM = localtime(&iTime);

	if (strftime(szBuffer, CTPP_ESCAPE_BUFFER_LEN, aArguments[0].GetString().c_str(), pTM) == 0) { return -1; }

	oCDTRetVal = szBuffer;

return 0;
}

//
// Get function name
//
CCHAR_P FnDateFormat::GetName() { return "date_format"; }

//
// A destructor
//
FnDateFormat::~FnDateFormat() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnBase64Encode
//

//
// Constructor
//
FnBase64Encode::FnBase64Encode()
{
	;;
}

//
// Handler
//
INT_32 FnBase64Encode::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Only 1 argument
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = Base64Encode(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnBase64Encode::GetName() { return "base64_encode"; }

//
// A destructor
//
FnBase64Encode::~FnBase64Encode() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnBase64Decode
//

//
// Constructor
//
FnBase64Decode::FnBase64Decode()
{
	;;
}

//
// Handler
//
INT_32 FnBase64Decode::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// Only 1 argument
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = Base64Decode(aArguments[0].GetString());

return 0;
}

//
// Get function name
//
CCHAR_P FnBase64Decode::GetName() { return "base64_decode"; }

//
// A destructor
//
FnBase64Decode::~FnBase64Decode() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnMD5
//
#ifdef MD5_SUPPORT

//
// Constructor
//
FnMD5::FnMD5()
{
	;;
}

//
// Handler
//
INT_32 FnMD5::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	static CHAR_8 aHex[] = "0123456789abcdef";
	CHAR_8 szMD5[32];
	// >0 arguments need
	if (iArgNum == 0) { return -1; }

	MD5_CTX oMD5Context;
	MD5_Init(&oMD5Context);

	INT_32 iI = iArgNum - 1;
	for (; iI >= 0; --iI)
	{
		// Data
		const std::string & sTMP = aArguments[iI].GetString();
		CCHAR_P szData = sTMP.c_str();
		// Data length
		const UINT_32 iDataLength = sTMP.size();

		MD5_Update(&oMD5Context, (const unsigned char *)szData, iDataLength);
	}

	unsigned char sDigest[16];
	MD5_Final(sDigest, &oMD5Context);

	INT_32 iJ = 0;
	for (iI = 0; iI < 16; ++iI)
	{
		szMD5[ iJ++] = aHex[ (sDigest[iI] >> 4 )& 0x0F ];
		szMD5[ iJ++] = aHex[  sDigest[iI]       & 0x0F ];
	}

	// Okay
	oCDTRetVal = std::string(szMD5, 32);

return 0;

}

//
// Get function name
//
CCHAR_P FnMD5::GetName() { return "md5"; }

//
// A destructor
//
FnMD5::~FnMD5() throw() { ;; }

#endif // MD5_SUPPORT

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnIconv
//
#ifdef ICONV_SUPPORT

//
// Constructor
//
FnIconv::FnIconv()
{
	;;
}

//
// Handler
//
INT_32 FnIconv::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// 3 arguments need
	if (iArgNum != 3) { return -1; }

	// Arg 2: destination charset
	// Arg 1: source charset
	// Arg 0: string to convert

	const std::string & sTo   = aArguments[2].GetString();
	const std::string & sFrom = aArguments[1].GetString();
	const std::string & sWhat = aArguments[0].GetString();

	std::string sFromTo(sFrom);
	sFromTo.append(sTo);

	iconv_t oIconvConverter = (iconv_t)(-1);

	std::map<std::string, iconv_t>::iterator itmIconvMap = mIconvMap.find(sFromTo);
	if (itmIconvMap != mIconvMap.end()) { oIconvConverter = itmIconvMap -> second; }
	// Try to open iconv converter
	else
	{
		oIconvConverter = iconv_open(sFrom.c_str(), sTo.c_str());

		if (oIconvConverter != (iconv_t)(-1)) { mIconvMap[sFromTo] = oIconvConverter; }
		else                                  { return -1;                            }
	}

	// Allocate memory
	size_t iSrcLength     = sWhat.size();
	size_t iDstLength     = iSrcLength / 2;
	for (;;)
	{

		char       * aDstData = new char[iDstLength];
#ifdef linux
		char       * aSrcData = (char *)sWhat.c_str();
#else
		const char * aSrcData = sWhat.c_str();
#endif
		char * aDstTMP       = aDstData;
		size_t iDstLengthTMP = iDstLength;
		size_t iResult       = iconv(oIconvConverter, &aSrcData, &iSrcLength, &aDstTMP, &iDstLengthTMP);

		if (iResult == (size_t)-1)
		{
			iconv(oIconvConverter, NULL, NULL, NULL, NULL);
			delete [] aDstData;
			if (errno != E2BIG) { return -1; }

			iDstLength *= 2;
			iSrcLength  = sWhat.size();
		}
		else
		{

			// Okay
			oCDTRetVal = std::string(aDstData, aDstTMP - aDstData);
			delete [] aDstData;
			break;
		}
	}

return 0;

}

//
// Get function name
//
CCHAR_P FnIconv::GetName() { return "iconv"; }

//
// A destructor
//
FnIconv::~FnIconv() throw()
{
	std::map<std::string, iconv_t>::iterator itmIconvMap = mIconvMap.begin();
	while (itmIconvMap != mIconvMap.end())
	{
		iconv_close(itmIconvMap -> second);
		++itmIconvMap;
	}
}

#endif // ICONV_SUPPORT

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnCast
//

//
// Constructor
//
FnCast::FnCast()
{
	;;
}

//
// Handler
//
INT_32 FnCast::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// 2 arguments need
	if (iArgNum != 2) { return -1; }

	// Arg 1: destination type
	// Arg 0: source type
	const std::string & sWhat = aArguments[1].GetString();
	if (sWhat.length() == 0) { return -1; }

	CDT & oTMP = aArguments[0];

	switch(sWhat[0])
	{
		// S[tring]
		case 'S':
		case 's':
			oCDTRetVal = oTMP.GetString();
			return 0;

		// O[ct[al]]
		case 'o':
		case 'O':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				unsigned long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%llo", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// D[ec[imal]]
		case 'd':
		case 'D':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%lld", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// I[nt[eger]]
		case 'i':
		case 'I':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%lli", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// H[ex[adecimal]]
		case 'H':
		case 'h':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetInt(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				unsigned long long iLL = 0;
				sscanf(oTMP.GetString().c_str(), "%llx", &iLL);
				oCDTRetVal = INT_64(iLL);
				return 0;
			}
			return -1;

		// F[loat]
		case 'F':
		case 'f':
			if      (oTMP.GetType() <= CDT::REAL_VAL)   { oCDTRetVal = oTMP.GetFloat(); return 0; }
			else if (oTMP.GetType() == CDT::STRING_VAL)
			{
				double dVal = 0;
				sscanf(oTMP.GetString().c_str(), "%lg", &dVal);
				oCDTRetVal = dVal;
				return 0;
			}
			return -1;

		// Error
		default:
			return -1;
	}
}

//
// Get function name
//
CCHAR_P FnCast::GetName() { return "cast"; }

//
// A destructor
//
FnCast::~FnCast() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnDefined
//

//
// Constructor
//
FnDefined::FnDefined()
{
	;;
}

//
// Handler
//
INT_32 FnDefined::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum == 0) { return -1; }

	oCDTRetVal = 1;

	INT_32 iI = iArgNum - 1;
	for (; iI >= 0; --iI)
	{
		if (aArguments[iI].GetType() == CDT::UNDEF)
		{
			oCDTRetVal = 0;
			break;
		}
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnDefined::GetName() { return "defined"; }

//
// A destructor
//
FnDefined::~FnDefined() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnSize
//

//
// Constructor
//
FnSize::FnSize()
{
	;;
}

//
// Handler
//
INT_32 FnSize::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum != 1) { return -1; }

	oCDTRetVal = aArguments[0].Size();

return 0;
}

//
// Get function name
//
CCHAR_P FnSize::GetName() { return "size"; }

//
// A destructor
//
FnSize::~FnSize() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnNumFormat
//

//
// Constructor
//
FnNumFormat::FnNumFormat()
{
	;;
}

//
// Handler
//
INT_32 FnNumFormat::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum != 2) { return -1; }

	std::string sTMP = aArguments[0].GetString();

	switch (aArguments[1].GetType())
	{
		// Integer value
		case CDT::INT_VAL:
			{
				CHAR_8 szRes[C_MAX_SPRINTF_LENGTH + 1];
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];

				szRes[C_MAX_SPRINTF_LENGTH] = '\0';

				INT_32 iCharacters = snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%lld", (long long)aArguments[1].GetInt());
				INT_32 iResPos = C_MAX_SPRINTF_LENGTH - 1;

				INT_32 iPos = 0;
				while (iCharacters >= 0)
				{
					szRes[iResPos] = szBuf[iCharacters];
					if (iPos++ % 3 == 0 && iPos != 1) { szRes[--iResPos] = sTMP[0]; }
					--iCharacters;
					--iResPos;
				}

				if (iPos % 3 == 1) { oCDTRetVal = &szRes[iResPos + 2]; }
				else               { oCDTRetVal = &szRes[iResPos + 1]; }
			}
			break;

		// Floating point value
		case CDT::REAL_VAL:
			{
				CHAR_8 szRes[C_MAX_SPRINTF_LENGTH + 1];
				CHAR_8 szBuf[C_MAX_SPRINTF_LENGTH + 1];

				szRes[C_MAX_SPRINTF_LENGTH] = '\0';

				INT_32 iCharacters = snprintf(szBuf, C_MAX_SPRINTF_LENGTH, "%.*G", CTPP_FLOAT_PRECISION, aArguments[1].GetFloat());
				INT_32 iResPos = C_MAX_SPRINTF_LENGTH - 1;

				INT_32 iTMP = iCharacters;
				iCharacters = 0;
				while (iCharacters < iTMP && szBuf[iCharacters++] != '.') { ;; }

				INT_32 iPos = 0;
				while (iTMP > iCharacters)
				{
					szRes[iResPos] = szBuf[iTMP];
					--iTMP;
					--iResPos;
				}

				while (iCharacters >= 0)
				{
					szRes[iResPos] = szBuf[iCharacters];
					if (iPos++ % 3 == 0 && iPos != 1) { szRes[--iResPos] = sTMP[0]; }
					--iCharacters;
					--iResPos;
				}

				if (iPos % 3 == 1) { oCDTRetVal = &szRes[iResPos + 2]; }
				else               { oCDTRetVal = &szRes[iResPos + 1]; }
			}
			break;

		// All other types
		default:
			oCDTRetVal = aArguments[1].GetString();
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnNumFormat::GetName() { return "num_format"; }

//
// A destructor
//
FnNumFormat::~FnNumFormat() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnObjDump
//

//
// Constructor
//
FnObjDump::FnObjDump(): pCDT(NULL)
{
	;;
}

//
// Pre-execution handler setup
//
INT_32 FnObjDump::PreExecuteSetup(OutputCollector          & oCollector,
                                  CDT                      & oCDT,
                                  const ReducedStaticText  & oSyscalls,
                                  const ReducedStaticData  & oStaticData,
                                  const ReducedStaticText  & oStaticText)
{
	pCDT = &oCDT;
return 0;
}

//
// Handler
//
INT_32 FnObjDump::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 0)
	{
		oCDTRetVal = "Global data object:\n";
		oCDTRetVal.Append(pCDT -> RecursiveDump());
	}
	else
	{
		oCDTRetVal = "";
		// Temp
		CHAR_8 szBuffer[CTPP_ESCAPE_BUFFER_LEN + 1];

		INT_32 iI = iArgNum - 1;
		for (; iI >= 0; --iI)
		{
			snprintf(szBuffer, CTPP_ESCAPE_BUFFER_LEN, "Argument %d:\n", iArgNum - iI);
			oCDTRetVal.Append(szBuffer);
			oCDTRetVal.Append(aArguments[iI].RecursiveDump());
		}
	}
return 0;
}

//
// Get function name
//
CCHAR_P FnObjDump::GetName() { return "obj_dump"; }

//
// A destructor
//
FnObjDump::~FnObjDump() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnVersion
//

//
// Constructor
//
FnVersion::FnVersion() { ;; }

//
// Handler
//
INT_32 FnVersion::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum == 1 && strcasecmp("full", aArguments[0].GetString().c_str()) == 0)
	{
		oCDTRetVal = "Engine: CTPP2 engine v" CTPP_VERSION " (" CTPP_IDENT "), copyright (c) 2004 - 2009 CTPP Dev. Team;\n"
		             "RuntimeLibrary: CTPP Standard Library v" CTPP_VERSION " (" CTPP_IDENT "), copyright (c) 2009 CTPP Dev. Team;\n"
		             "License: BSD-like, see " CTPP_MASTER_SITE_URL ";";
	}
	else
	{
		oCDTRetVal = "CTPP2 engine v" CTPP_VERSION " (" CTPP_IDENT "), copyright (c) 2004 - 2009 CTPP Dev. Team";
	}

return 0;
}

//
// Get function name
//
CCHAR_P FnVersion::GetName() { return "version"; }

//
// A destructor
//
FnVersion::~FnVersion() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnError
//

//
// Constructor
//
FnError::FnError()
{
	;;
}

//
// Handler
//
INT_32 FnError::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal) { return -1; }

//
// Get function name
//
CCHAR_P FnError::GetName() { return "error"; }

//
// A destructor
//
FnError::~FnError() throw()
{
	;;
}

// //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnBBCode
//

//
// Constructor
//
FnBBCode::FnBBCode() { ;; }

//
// Pre-execution handler setup
//
INT_32 FnBBCode::PreExecuteSetup(OutputCollector          & oCollector,
                                 CDT                      & oCDT,
                                 const ReducedStaticText  & oSyscalls,
                                 const ReducedStaticData  & oStaticData,
                                 const ReducedStaticText  & oStaticText)
{
	//
return 0;
}

//
// Handler
//
INT_32 FnBBCode::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	std::string sTMP(aArguments[0].GetString());

	//
	// [b]
	// [i]
	// [u]
	// [s]
	// [quote name="username" post="123"]
	// [size]
	// [code]
	// [color]
	// [img]
	// [url]
	// [youtube]
	// [rutube]
	// [right]
	// [left]
	// [center]
	// [spoiler]
	// [topic]
	// [post]
	// [forum]
	//

return 0;
}

//
// Get function name
//
CCHAR_P FnBBCode::GetName() { return "bb_code"; }

//
// A destructor
//
FnBBCode::~FnBBCode() throw() { ;; }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnDefault
//

//
// Constructor
//
FnDefault::FnDefault()
{
	;;
}

//
// Handler
//
INT_32 FnDefault::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	// At least one argument need
	if (iArgNum != 2) { return -1; }

	// Get first arument
	oCDTRetVal = aArguments[1];

	// Set default value from second argument
	if (aArguments[1].GetType() == CDT::UNDEF) { oCDTRetVal = aArguments[0]; }
	else                                       { oCDTRetVal = aArguments[1]; }

return 0;
}

//
// Get function name
//
CCHAR_P FnDefault::GetName() { return "default"; }

//
// A destructor
//
FnDefault::~FnDefault() throw()
{
	;;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Class FnJSON
//

//
// Constructor
//
FnJSON::FnJSON()
{
	;;
}

//
// Handler
//
INT_32 FnJSON::Handler(CDT * aArguments, const UINT_32 & iArgNum, CDT & oCDTRetVal)
{
	if (iArgNum != 1) { return -1; }

	std::string sData;

	CDT2JSON(aArguments[0], sData);

	oCDTRetVal = sData;

return 0;
}

//
// Get function name
//
CCHAR_P FnJSON::GetName() { return "JSON"; }

//
// A destructor
//
FnJSON::~FnJSON() throw() { ;; }

} // namespace CTPP
// End.
